const express = require('express');
const cwd = process.cwd();
const os = require('os');
const helmet = require('helmet');
const cookieParser = require('cookie-parser');
const bodyParser = require('body-parser');
const { CacheControl, I18nProxy } = require('./lib');
const accetLanguageParser = require('accept-language-parser');
const Logger = require('ocbesbn-logger');
const extend = require('extend');
const ServiceClient = require('ocbesbn-service-client');
const UserIdentity = require('@opuscapita/useridentity-middleware');
const Bouncer = require('@opuscapita/bouncer');
const Promise = require('bluebird');
const pathjs = require('path');
const { I18nManager } = require('@opuscapita/i18n');
const fs = require('fs');
const formatters = require('./i18n/formatters');
const EventClient = require('@opuscapita/event-client');
const si = require('systeminformation');

/**
 * Module for general and common initialization of the OpusCapita Business Network
 * web server module providing default behavior for RESTful services.
 * @module @opuscapita/web-init
 */

 /**
  * Object describing an endpoint API to be registered with the API gateway.
  * @typedef {object} ApiEndpoint
  * @property {string} publicRoute URI for the public facing path of an endpoint e.g. /user/users.
  * @property {string} serviceRoute URI for the service path of an endpoint.
  */

/**
 * Static object containing enumeration objects providing server behavior configuration.
 */
module.exports.Server = {
    /**
     * Enumeration made to control a server's environment behavior.
     * @enum {number}
     */
    Mode : { None : 0, Productive : 1, Dev : 2, Debug : 3 },

    /**
     * List of flags for security configuration.
     * @enum {number}
     */
    Security : { All : 0, AllowCrossOrigin : 1 } // Bitword
}

/**
 * Contains the name of the current service this module is running in. It simply takes the last directory
 * name of the main processes path.
 */
module.exports.serviceName = cwd.slice(cwd.lastIndexOf('/') + 1);

/**
 * Contains the name of the host the code is running on.
 */
module.exports.hostname = os.hostname();

/**
 * Static object representing a default configuration set for web-init. Have a look at the [actual values]{@link DefaultConfig}.
 *
 * @property {object} server - All server web server related settings.
 * @property {Server.Mode} server.mode - Environment [mode]{@link module:@opuscapita/web-init.Server.Mode} to run in.
 * @property {Server.Security} server.security - Bitword for applying multiple [security]{@link module:@opuscapita/web-init.Server.Security} presets.
 * @property {array} server.crossOrigins - A list of allowed cross-origins source domains.
 * @property {number} server.maxBodySize - Maximum size in bytes of a http-reuqest-body.
 * @property {string} server.staticFilePath - Path for all static files delivered by the web server.
 * @property {string} server.indexFilePath - Path of a file delivered by the web server on client requests to. /.
 * @property {array} server.indexFileRoutes - List of routes the index file should be delivered on.
 * @property {string} server.hostname - The local host name or ip-address to bind the server to.
 * @property {number} server.port - The local tcp port for the server to listen on.
 * @property {object} server.events - Object providing access points to some life cycle events.
 * @property {function} server.events.onStart - Start event fired when successfully listens for incoming connections.
 * @property {function} server.events.onEnd - End event fired whenever the server shuts down.
 * @property {function} server.events.onRequest - Fired every time, a request comes in.
 * @property {function} server.events.onError - Fired if the server could not start.
 * @property {object} server.webpack - Webpack configuration.
 * @property {boolean} server.webpack.useWebpack - Controls whenever the server should use Webpack.
 * @property {boolean} server.webpack.configFilePath - Webpack configuration file path.
 * @property {array} server.middlewares - Array of additional middleware objects to be used with the express web server.
 * @property {array} server.requestHeadersToResponse - Array of request headers to be copied to the response.
 * @property {array} server.appVariables - Object containing additional variables to be passed to express.
 * @property {array} server.enableIdentity - Enables or disables the use of the [useridentity-middleware]{@link https://github.com/OpusCapita/useridentity-middleware}.
 * @property {array} server.enableBouncer - Enables or disables the use of the [bouncer]{@link https://github.com/OpusCapita/bouncer} middleware.
 * @property {array} server.bouncer - Object containing additional configurations for the [bouncer]{@link https://github.com/OpusCapita/bouncer} middleware.
 * @property {array} server.enableEventClient - Injects a prepared instance of [EventClient]{@link https://github.com/OpusCapita/event-client} into the req.opuscapita namespace.
 * @property {object} server.cacheControl - Object containing additional configurations to control http response caching.
 * @property {number} server.cacheControl.defaultAge - Default cache timeout in seconds.
 * @property {array} server.cacheControl.cachableEndpoints - List of regexp objects defining cachable endpoints.
 * @property {array} server.cacheControl.nonCachableEndpoints - List of regexp objects defining non-cachable endpoints.
 * @property {ApiEndpoint[]} server.publicApis - List of api endpoints to be registered for the service with the API gateway.
 * @property {array} server.noLogEndpoints - List of api endpoints that should not be logged in order to reduce log output. Will be ignored if Server.Mode is Dev.
 * @property {Logger} logger - Logger instance to be used by web-init.
 * @property {object} serviceClient - Configuration for injecting a [ServiceClient]{@link https://github.com/OpusCapitaBusinessNetwork/service-client} instance into every request.
 * @property {boolean} serviceClient.consul - Configuration options for service discovery done by the ServiceClient.
 * @property {boolean} serviceClient.consul.host - Hostname for a Consul service discovery server.
 * @property {object} serviceClient.caching - Configuration options for service-client request caching.
 * @property {string} serviceClient.caching.driver - Cache driver to use.
 * @property {number} serviceClient.caching.defaultExpire - Global cache entry expiration time in seconds.
 * @property {array} serviceClient.headersToProxy - List of source headers to be copied to the context of a [ServiceClient]{@link https://github.com/OpusCapitaBusinessNetwork/service-client} instance. Headers starting with X- are always added.
 * @property {object} routes - Basic routing configuration for the RESTful web server.
 * @property {boolean} routes.addRoutes - Controls whenever routes should be added to be accessible via http.
 * @property {array} routes.modulePaths - List of modules to load in order to register wev server routes.
 * @property {object} routes.dbInstance - Databse object to be passed to every module registering routes.
 */
module.exports.DefaultConfig = {
    server : {
        mode : process.env.NODE_ENV === 'development' ? this.Server.Mode.Dev : this.Server.Mode.Productive,
        security : this.Server.Security.All,
        crossOrigins : [ '*' ],
        maxBodySize : 1048576, // 1 MiB
        staticFilePath : express.static(__dirname + '/static'),
        indexFilePath : null,
        indexFileRoutes : [ '/' ],
        hostname : process.env.HOST || '0.0.0.0',
        port : process.env.PORT || 3000,
        events : {
            onStart : function(server) { },
            onEnd : function(server) { },
            onRequest : function(req, res, next) { next(); },
            onError : function(err, server) { process.stderr.write(JSON.stringify(err)); }
        },
        webpack : {
            useWebpack : false,
            configFilePath : process.cwd() + '/webpack.conf.js'
        },
        middlewares : [ ],
        requestHeadersToResponse : [ 'Correlation-Id' ],
        appVariables : { },
        enableIdentity : true,
        enableBouncer : true,
        bouncer : {
            registerPermissions : {
                retryTimeout : 1000,
                retryCount : 30
            }
        },
        enableEventClient : true,
        cacheControl : {
            defaultAge : 3600,
            cachableEndpoints : [ '^/static/*', '^/public/static/*' ],
            nonCachableEndpoints : [ '^/api/*', '^/public/api/*' ]
        },
        publicApis : [ '*' ],
        noLogEndpoints : [ '^/api/list/apis$', '^/api/health/check$', '^/api/health/stats$' ]
    },
    logger : new Logger(),
    serviceClient : {
        consul : {
            host : 'consul'
        },
        caching : {
            driver : 'dummy',
            defaultExpire : 600
        },
        headersToProxy : [ 'Cookie', 'Authorization', 'From', 'Correlation-Id' ]
    },
    routes : {
        addRoutes : true,
        modulePaths : [ process.cwd() + '/src/server/routes' ],
        dbInstance : null
    }
}

module.exports.getHealthStatus = async function()
{
    let isHealthy = true;

    for(const module of this.allRouteModules)
    {
        if(typeof module.onHealthCheck === 'function')
            isHealthy = isHealthy && await module.onHealthCheck();
    }

    return isHealthy;
}

module.exports.getSystemMetrics = async function()
{
    return {
        cpuTemperature : await si.cpuTemperature(),
        cpuLoad : await si.currentLoad(),
        memory : await si.mem(),
        fsSize : await si.fsSize(),
        fsStats : await si.fsStats(),
        //diskIO : await si.disksIO(),
        //networkStats : await si.networkStats()
    };
}

/**
 * Initializes an instance of web-init by passing a set of optional configuration values.
 * This method also registeres the process signals SIGTERM and SIGINT and runs a clean shutdown on the server.
 *
 * @param {object} config - Optional configuration object extending {@link module:@opuscapita/web-init.DefaultConfig}.
 * @returns {Promise} [Promise]{@link http://bluebirdjs.com/docs/api-reference.html} containing an [Express]{@link https://github.com/expressjs/express} instance.
 */
module.exports.init = function(config)
{
    this.config = config = extend(true, { }, this.DefaultConfig, config);
    config.serviceClient.headersToProxy = config.serviceClient.headersToProxy.map(item => item.toLowerCase());

    this.allRouteModules = [ ];

    const logger = config.logger;
    this.logger = logger;

    logger.info('Starting up web server... Host: %s, Port: %s', config.server.hostname, config.server.port);

    if(typeof config.server.staticFilePath === 'string')
        config.server.staticFilePath = express.static(config.server.staticFilePath);

    const app = express();
    this.app = app;

    for(const key in config.server.appVariables)
        app.set(key, config.server.appVariables[key]);

    app.use((req, res, next) =>
    {
        req.opuscapita = req.opuscapita || { };
        req.opuscapita.timerStart = process.hrtime();

        const languages = req.headers["accept-language"] || 'en';
        req.opuscapita.acceptLanguage = accetLanguageParser.parse(languages);

        req.opuscapita.logger = new Logger({
            context : {
                serviceName : this.serviceName,
                method : req.method,
                requestUri : req.originalUrl,
                correlationId : req.headers['correlation-id']
            }
        });

        res.on('finish', async () =>
        {
            req.opuscapita._eventClient && await req.opuscapita._eventClient.dispose().catch(e => null);

            const doLog = config.server.noLogEndpoints.reduce((all, val) => all && !req.originalUrl.match(val), true);

            if(doLog || config.server.mode === this.Server.Mode.Dev)
            {
                const timerRun = process.hrtime(req.opuscapita.timerStart);
                const timeTaken = (timerRun[0] * 1000) + (timerRun[1] / 1e6);

                req.opuscapita.logger.info(`Request for ${req.originalUrl} finished in ${Math.round(timeTaken * 100) / 100} ms.`);
            }
        });

        next();
    });

    app.use(helmet());
    app.use(cookieParser());
    app.use(bodyParser.json({ limit : config.server.maxBodySize }));
    app.use(bodyParser.urlencoded({ extended: false, limit : config.server.maxBodySize }));
    app.use(new CacheControl(config.server.cacheControl).getMiddleware());

    app.use((req, res, next) =>
    {
        const serviceName = this.serviceName;

        Object.defineProperty(req.opuscapita, 'serviceClient',
        {
            get : function()
            {
                if(!this._serviceClient)
                {
                    const headersToProxy = config.serviceClient.headersToProxy;
                    const localHeaders = { };

                    for(const key in req.headers)
                    {
                        if(key.startsWith('x-') || headersToProxy.indexOf(key) !== -1)
                            localHeaders[key] = req.headers[key];
                    }

                    localHeaders['x-origin-service'] = serviceName;

                    const serviceClientConfig = extend(true, { }, config.serviceClient, { logger : req.opuscapita.logger.clone() });

                    this._serviceClient = new ServiceClient(serviceClientConfig);
                    this._serviceClient.contextify({ headers : localHeaders });
                }

                return this._serviceClient;
            },
            set : function()
            {
                throw new Error('You are not allowed to set this property.');
            }
        });

        next();
    });

    if(config.server.enableIdentity)
    {
        const userIdentity = new UserIdentity({ noLogEndpoints : config.server.noLogEndpoints });
        app.use(userIdentity.middleware());

        app.use((req, res, next) =>
        {
            req.opuscapita.logger.contextify({
                userId : req.opuscapita.userData('id')
            });

            next();
        });
    }

    app.use((req, res, next) =>
    {
        const serviceName = this.serviceName;

        Object.defineProperty(req.opuscapita, 'eventClient',
        {
            get : function()
            {
                if(!this._eventClient)
                {
                    const context = {
                        serviceName : serviceName,
                        method : req.method,
                        requestUri : req.originalUrl,
                        correlationId : req.headers['correlation-id'],
                        userId : req.opuscapita.userData && req.opuscapita.userData('id')
                    };

                    this._eventClient = new EventClient();
                    this._eventClient.contextify(context);
                }

                return this._eventClient;
            },
            set : function()
            {
                throw new Error('You are not allowed to set this property.');
            }
        });

        next();
    });

    app.use((req, res, next) =>
    {
        const acceptLang = req.opuscapita.acceptLanguage[0];
        const locale = (req.opuscapita.userData && req.opuscapita.userData('languageId'))
            || acceptLang && acceptLang.code || 'en';

        const manager = new I18nManager({ locale, fallbackLocale : 'en', localeFormattingInfo : formatters });
        req.opuscapita.i18n = new I18nProxy(manager);

        if(config.routes.modulePaths)
        {
            config.routes.modulePaths.forEach(path =>
            {
                const translationsPath = pathjs.join(pathjs.resolve(path),  'i18n');

                if(fs.existsSync(translationsPath))
                    manager.register(translationsPath, require(translationsPath));
            });
        }

        next();
    });

    app.use((req, res, next) =>
    {
        config.server.requestHeadersToResponse.map(key => res.set(key, req.headers[key.toLowerCase()]));

        res.set('X-Service-Name', this.serviceName);
        res.set('X-Service-Host', this.hostname);

        const doLog = config.server.noLogEndpoints.reduce((all, val) => all && !req.originalUrl.match(val), true);

        if(doLog || config.server.mode === this.Server.Mode.Dev)
            req.opuscapita.logger.info(`Incoming request for: ${req.originalUrl}`);

        next();
    });

    if(config.server.enableEventClient)
    {
        const eventClient = new EventClient();
        eventClient.init().catch(e => null).then(() => eventClient.dispose()).catch(e => null);
    }

    if(config.server.enableBouncer)
    {
        const bouncer = new Bouncer(config.server.bouncer);
        bouncer.registerPermissions(config.server.bouncer.registerPermissions).catch(e => logger.error(e));

        app.use(bouncer.middleware());
    }

    if(config.server.middlewares)
        config.server.middlewares.forEach(obj => app.use(obj));

    if(config.server.security && this.Server.Security.AllowCrossOrigin)
    {
        logger.info('Allowing cross origin requests for: %j', config.server.crossOrigins);

        app.use((req, res, next) =>
        {
        	res.header('Access-Control-Allow-Origin', config.server.crossOrigins.join(' ')); // TODO: Only trusted hosts.
        	res.header('Access-Control-Allow-Headers', 'Origin, X-Requested-With, Content-Type, Accept');
            res.header('Vary', 'Origin');

        	next();
        });
    }

    if(config.server.staticFilePath)
        app.use('/static', config.server.staticFilePath);

    if(config.server.mode === this.Server.Mode.Productive)
    {
        if(config.server.indexFilePath)
        {
            const fullIndexPath = pathjs.resolve(config.server.indexFilePath);
            app.get(config.server.indexFileRoutes, (req, res) => res.sendFile(fullIndexPath));
        }
    }
    else
    {
        logger.info('Using webpack.');

        if(config.server.webpack.useWebpack)
        {
            if(process.env.NODE_ENV !== 'development')
                console.warn('You\'re trying to use webpack outside of the development environment.');

            const webpack = require('webpack');
            const webpackMiddleware = require('webpack-dev-middleware');
            const webpackConfig = require(config.server.webpack.configFilePath);
            const webpackCompiler = webpack(webpackConfig);
            const middleware = webpackMiddleware(webpackCompiler, {
                publicPath : webpackConfig.output && webpackConfig.output.publicPath,
                noInfo : false,
                watchOptions : {
                    aggregateTimeout : 300,
                    poll : true
                }
            });

            app.use(middleware);

            if(config.server.indexFilePath)
            {
                app.get(config.server.indexFileRoutes, (req, res) =>
                {
                    const absolutePath = pathjs.resolve(config.server.indexFilePath);

                    if(fs.existsSync(absolutePath))
                    {
                        res.set('Content-Type', 'text/html');
                        res.sendFile(pathjs.resolve(config.server.indexFilePath));
                    }
                    else
                    {
                        res.status(404).json({ message : 'The requested file does not exist.' });
                    }
                });
            }
        }
    }

    app.use(config.server.events.onRequest);

    app.get('/api/health/check', async (req, res) =>
    {
        if(await this.getHealthStatus())
            res.json({ message : "Yes, I'm alive!" });
        else
            res.status(503).json({ message : "He's dead Jim!" });
    });

    app.get('/api/health/stats', async (req, res) =>
    {
        res.json(await this.getSystemMetrics());
    });
    
    app.get('/api/list/apis', (req, res) =>
    {
        if(Array.isArray(config.server.publicApis))
        {
            const apisToBeSent = [ ];

            if(config.server.publicApis.includes('*'))
            {
                apisToBeSent.push({ publicRoute : `/${this.serviceName}`, serviceRoute : `/${this.serviceName}` });
            }
            else
            {
                for(const api of config.server.publicApis)
                {
                    if(api.publicRoute && api.serviceRoute)
                        apisToBeSent.push({ publicRoute : api.publicRoute, serviceRoute : api.serviceRoute });
                }
            }

            if(apisToBeSent.length)
                res.json(apisToBeSent)
            else
                res.status(204).end();
        }
        else
        {
            res.status(204).end();
        }
    });

    let preCondition;

    if(config.routes.addRoutes === true)
    {
        logger.info('Adding routes...');

        const runInit = (path, app, db, config) =>
        {
            const module = require(path);
            this.allRouteModules.push(module);

            return module.init(app, db, config);
        }

        const inits = ((paths, db, config) => Array.isArray(paths) ? paths.map(path => runInit(path, app, db, config))
            : [ runInit(paths, app, db, config) ])(config.routes.modulePaths, config.routes.dbInstance, config.routes);

        preCondition = Promise.all(inits).then(() => logger.info('Routes initialized!'));
    }
    else
    {
         preCondition = Promise.resolve()
    }

    return preCondition.then(() =>
    {
        app.use((err, req, res, next) =>
        {
            req.opuscapita.logger.error(err.stack);
            res.status(500).send({ message : err.message });
        });

        const onServerListen = () => config.server.events.onStart.call(this, app);
        const onServerError = (err) => { config.server.events.onError.call(this, err, app); this.end(); }
        const onServerEnd = () => config.server.events.onEnd.call(this, app);

        this.server = app.listen(config.server.port, config.server.hostname, 1000, onServerListen);
        this.server.on('error', onServerError);
        this.server.on('close', onServerEnd);

        process.on('SIGTERM', () => { this.end(); process.exit(); });
        process.on('SIGINT', () => { this.end(); process.exit(); });

        logger.info('Server started.');

        return app;
    })
    .catch(error =>
    {
        logger.error(error.message);

        return this.end().catch(e => null).then(() => { throw error; });
    });
}

/**
 * Ends an initialized instance of web-init.
 * @returns {Promise} [Promise]{@link http://bluebirdjs.com/docs/api-reference.html}.
 */
module.exports.end = function()
{
    return new Promise(async (resolve, reject) =>
    {
        try
        {
            if(Array.isArray(this.allRouteModules))
                await Promise.all(this.allRouteModules.map(module => module.end && module.end())).catch(e => this.logger.error(e)).catch(e => null);

            if(this.server)
            {
                this.server.close(() =>
                {
                    this.config.server.events.onEnd.call(this, this.app);
                    this.server = null;

                    resolve();
                });
            }
            else
            {
                resolve();
            }
        }
        catch(e)
        {
            reject(e);
        }
    });
}
