const extend = require('extend');

class CacheControl
{
    constructor(config)
    {
        this.config = extend(true, { }, CacheControl.DefaultConfig, config);
    }

    getMiddleware()
    {
        return (req, res, next) =>
        {
            const oldSet = res.set.bind(res);

            const safeSet = (...args) =>
            {
                if(res.headersSent)
                    req.opuscapita.logger.warn('Headers have already been sent by output. Refusing header action.');
                else
                    return oldSet(...args);
            }

            res.opuscapita = res.opuscapita || { };
            res.opuscapita.setMaxAge = (seconds) =>
            {
                if(seconds > -1)
                {
                    safeSet('Cache-Control', `public, max-age=${seconds}, s-maxage=${seconds}`);
                    safeSet('Expires', new Date(new Date().getTime() + (seconds * 1000)).toUTCString());
                    safeSet('Vary', 'ETag, Expires');

                    res.removeHeader('Pragma');
                }
                else
                {
                    safeSet('Cache-Control', 'public, no-cache, must-revalidate, max-age=0, proxy-revalidate, s-maxage=0');
                    safeSet('Pragma', 'no-cache');
                    safeSet('Expires', '-1');
                    safeSet('Vary', '*');
                }

                return res;
            }

            res.opuscapita.setNoCache = () => res.opuscapita.setMaxAge(-1);

            res.set = (name, value) =>
            {
                if(name && name.toLowerCase() === 'expires')
                {
                    let seconds = parseInt(value);

                    if(isNaN(seconds))
                        seconds = Math.round((new Date(value) - new Date()) / 1000);

                    return res.opuscapita.setMaxAge(seconds);
                }
                else
                {
                    return safeSet(name, value);
                }
            }

            if(process.env.NODE_ENV === 'development')
            {
                res.opuscapita.setMaxAge(-1);
            }
            else
            {
                const epIsCachable = this.config.cachableEndpoints.filter(ep => req.url.match(new RegExp(ep))).length > 0;
                const epIsNotCachable = this.config.nonCachableEndpoints.filter(ep => req.url.match(new RegExp(ep))).length > 0;

                if(epIsCachable && !epIsNotCachable && typeof(this.config.defaultAge) !== 'undefined')
                    res.opuscapita.setMaxAge(this.config.defaultAge);
                else if(epIsNotCachable)
                    res.opuscapita.setMaxAge(-1);
            }

            next();
        }
    }
}

CacheControl.DefaultConfig = {
    defaultAge : -1,
    cachableEndpoints : [ ],
    nonCachableEndpoints : [ ]
}

module.exports = CacheControl;
